-- Puffin: A Tojaso FowlPlay Project
-- View component for a bar group that shows timer bars for all spells in a list.
-- Each bar is assembled from a collection of widgets, as defined by a layout.
-- An anchor frame is constructed and icons are displayed relative to it.

-- View API:
-- View_BarGroup.Create() - a new instance of a bar group view
-- View_BarGroup:Release() - release an instance of a bar group view, recycling it for future use
-- View_BarGroup:Configure() - configure based on current defaults and widget descriptions
-- View_BarGroup:Update() - update the bar group

-- View attributes:
-- anchorFrame - frame to anchor the icon group to (default: UIParent) [c]
-- anchorPoint - if anchor defined then point to position relative to (default: BOTTOMLEFT) [c]
-- relativePoint - point on this widget to position relative to (default: BOTTOMLEFT) [c]
-- offsetX - horizontal offset from anchor [c]
-- offsetY - vertical offset from anchor [c]
-- layout - table of attributes used to configure the icon group
-- sfx - animations, color changes, start and expiration sounds
-- depth - display depth for this view [c]
-- scale - factor to scale view's display size [d]
-- alpha - alpha for the icon group [d]
-- hidden - hide the view [d]

-- Layout attributes:
-- defaults - table with default settings, including those for each class of widget [c]
-- button - widget template for an action button [c]
-- label - widget template for label text field [c]
-- timer - widget template for timer text field [c]
-- count - widget template for count text field [c]
-- bar - widget template for the timer bar [c]
-- backdrop - widget template for a background panel that dynamically scales with active bars [c]
-- direction - "up", "down" (default: "up") [c]
-- barDirection -- "right-to-left" or "left-to-right" (default "right-to-left") [c]
-- fillBars - if true then fill instead of empty bars (default: false) [c]
-- limit - maximum number of bars to show (default: 10) [c]
-- iconOffsetX - distance between button and bar widgets (default: 0) [c]
-- spacingY - vertical distance between bars (default: 0) [c]
-- wrap - how many bars to show before wrapping to next row or column (default: 0, i.e., don't wrap) [c]
-- wrapDirection - true for "left" or "up", false for "right" or "down" [c]

local View_BarGroup = {}
local EF = Puffin.ErrorFormat
local DB, VC
local classDepth = 0 -- mininum depth for view's of this class
local cc = { a = 1 } -- temporary color table with alpha preset

-- Return an initialized instance of the view
function View_BarGroup.Create() return VC.CreateView() end

-- Release a view when no longer needed, recycle after clearing attributes and hiding widgets associated with it
function View_BarGroup:Release() VC.ReleaseView(self) end

-- Configure bar (with button, bar, label text, timer text, and count text widgets) associated with this view
function View_BarGroup:Configure()
	assert(self.defaults, EF("View_BarGroup", "undefined defaults"))
	assert(self.layout, EF("View_BarGroup", "undefined layout"))
	local offsetDepth = ((self.depth or 0) + classDepth) * 10
	local xname = string.gsub(self.name, " ", "_")
	self.frame = CreateFrame("Frame", "PuffinBarGroup" .. xname, UIParent) -- create an anchor frame
	self.frame:SetFrameLevel(self.frame:GetFrameLevel() + 20) -- set frame level higher than other addons
	
	self.frame:SetSize(1, 1) -- one pixel size for the anchor frame
	self.frame:SetPoint(self.anchorPoint or "BOTTOMLEFT", self.anchorFrame or UIParent, self.relativePoint or "BOTTOMLEFT",
		self.offsetX or 0, self.offsetY or 0)
	self.frame:Show()

	-- create and configure the widgets for the maximum number of bars in the bar group
	local layout = self.layout
	assert(layout.button and layout.bar and layout.label and layout.timer and layout.count, EF("View_IconGroup", "incomplete layout")) -- required fields
	self.bars = Puffin.AllocateTable()
	for i = 1, (layout.limit or 10) do
		local bar = Puffin.AllocateTable()
		local t = VC.CreateWidget(layout.button)
		if t then
			bar.button = t; t.anchor = self.frame; t.parent = self.frame; t.masque = self.name
			t.onEnter = self.onEnter; t.onLeave = self.onLeave; t.onClick = self.onClick
			VC.ConfigureWidget(self, t, layout.button, offsetDepth)
		end
		t = VC.CreateWidget(layout.bar)
		if t then
			bar.bar = t; t.anchor = bar.button; t.parent = self.frame
			t.onEnter = self.onEnter; t.onLeave = self.onLeave; t.onClick = self.onClick
			VC.ConfigureWidget(self, t, layout.bar, offsetDepth)
		end
		t = VC.CreateWidget(layout.label)
		if t then bar.label = t; t.anchor = bar.bar; t.parent = bar.button; VC.ConfigureWidget(self, t, layout.label, offsetDepth) end	
		t = VC.CreateWidget(layout.timer)
		if t then bar.timer = t; t.anchor = bar.bar; t.parent = bar.button; VC.ConfigureWidget(self, t, layout.timer, offsetDepth) end	
		t = VC.CreateWidget(layout.count)
		if t then bar.count = t; t.anchor = bar.button; t.parent = bar.button; VC.ConfigureWidget(self, t, layout.count, offsetDepth) end
		self.bars[i] = bar
	end

	if layout.backdrop then self.backdrop = VC.CreateBackdrop(layout.backdrop) end -- create display panel widget for optional backdrop
end

-- Update all widgets associated with this view, using widget-specific providers to access the data record
function View_BarGroup:Update(spellCount, spellList)
	local layout = self.layout
	assert(layout, EF("View_BarGroup", "undefined layout"))
	local wrap = 0 -- indicates default of not wrapping
	local dir = layout.direction and 1 or -1 -- will be 1 or -1 depending on growth direction
	local dx, dy = dir * (layout.width + layout.spacingX), dir * (layout.height + layout.spacingY) -- offset to each new bar
	local wx, wy = 0, 0 -- offsets from starting point when need to wrap
	local bw, bh = 0, 0 -- number of bar widths and heights in size of backdrop
	local xoffset, yoffset, xdir, ydir, wadjust = 0, 0, 1, dir, 0 -- position adjustments for backdrop
	local count, maxBars, cdir = spellCount, layout.limit, 0
	if not maxBars or (maxBars == 0) then maxBars = count end
	if count > maxBars then count = maxBars end
	local ac = count -- actual count before wrap adjustment
	if layout.wrap then wrap = layout.wrap; if (wrap > 0) and (count > wrap) then count = wrap end end
	local anchorPoint = "BOTTOMLEFT"
	wx = -dx; dx = 0; if count > 0 then bw = math.ceil(ac / count) else bw = 1 end; bh = count; wadjust = layout.iconOffsetX
	if layout.wrapDirection then xoffset = layout.iconOffsetX else xoffset = layout.iconOffsetX - (bw - 1) * (layout.width + layout.spacingX) end
	if not layout.direction then anchorPoint = "TOPLEFT"; wx = -wx end
	if layout.wrapDirection then wx = -wx end
	if wrap > 0 then -- attachment options differ for wrapping bar groups
		self.lastRow = wx * bw; self.lastColumn = dy * count; self.lastX = nil; self.lastY = nil
	else
		self.lastX = 0; self.lastY = dy * count; self.lastRow = nil; self.lastColumn = nil
	end

	local anchorFrame = self.frame
	anchorFrame:SetAlpha(self.alpha or 1); anchorFrame:SetScale(self.scale or 1)
	self.count = 0
	for i = 1, layout.limit do
		local bar = self.bars[i]
		if i <= spellCount then
			local w, skip = i - 1, 0; if wrap > 0 then skip = floor(w / wrap); w = w % wrap end
			local x1 = (dx * w) + (wx * skip); local y1 = (dy * w) + (wy * skip)
			local s = spellList[i]
			local t = bar.button
			self.count = self.count + 1
			t._class.SetPosition(t, anchorPoint, anchorFrame, anchorPoint, x1, y1)
			if self.sfx then
				if self.sfx.clockOverlay and s.duration > 0 then
					t.clockExpire = s.expire; t.clockDuration = s.duration; t.clockDirection = self.sfx.clockDirection
				else
					t.clockDuration = nil
				end
				if self.sfx.masqueColor then cc.r = s.r; cc.g = s.g; cc.b = s.b; t.masqueColor = cc else t.masqueColor = nil end
			end
			local e = t.eventHandler; if e then e.source = s.source; e.tooltip = s.tooltip; e.unit = s.unit; e.caster = s.caster end
			t.hidden = false; t.texture = s.icon; t._class.Update(t)
			t = bar.bar; t.hidden = false; t.expire = s.expire; t.duration = s.duration; t.maxTime = s.maxTime
			cc.r = s.r; cc.g = s.g; cc.b = s.b; t.fgColor = cc; t.bgColor = cc; t._class.Update(t)
			t = bar.label; t.hidden = false; t.text = s.label or s.name; t._class.Update(t)
			t = bar.timer; t.hidden = false; t.timeMax = s.duration; t.timeExpire = s.expire; t._class.Update(t)
			t = bar.count; if (s.count or 0) > 1 then t.hidden = false; t.text = tostring(s.count) else t.hidden = true; t.text = "" end; t._class.Update(t)
		else
			local t = bar.button; t.hidden = true; table.wipe(t.eventHandler); t._class.Update(t)
			t = bar.bar; t.hidden = true; t._class.Update(t)
			t = bar.label; t.hidden = true; t._class.Update(t)
			t = bar.timer; t.hidden = true; t._class.Update(t)
			t = bar.count; t.hidden = true; t._class.Update(t)
		end
	end

	local t = self.backdrop -- optional backdrop adjusted after icons are arranged
	if t then
		if spellCount > 0 then
			local offsetDepth = ((self.depth or 0) + classDepth) * 10
			t.width = bw * layout.width; if bw > 1 then t.width = t.width + ((bw - 1) * layout.spacingX) end
			t.height = bh * layout.height; if bh > 1 then t.height = t.height + ((bh - 1) * layout.spacingY) end
			t.hidden = false; t.anchorPoint = anchorPoint; t.relativePoint = anchorPoint; t.anchor = anchorFrame; t.offsetX = xoffset; t.offsetY = yoffset
			VC.ConfigureBackdrop(self, t, layout.backdrop, offsetDepth)
		else
			t.hidden = true
		end
		t._class.Update(t)
	end
end

function View_BarGroup.OnEnable()
	DB = Puffin.Component("Data") -- link to data management functions
	assert(DB, EF("View_BarGroup", "Data component not found"))
	VC = Puffin.Component("View_Class") -- link to view class functions
	assert(VC, EF("View_BarGroup", "View class component not found"))
end

Puffin.RegisterComponent("View_BarGroup", View_BarGroup, "VIEW", 2)
