-- Puffin: A Tojaso FowlPlay Project
-- View component for a data broker-compatible plugin to display with a single icon and single text field.

-- View API:
-- View_BrokerBlock.Create() - a new instance of a broker block view
-- View_BrokerBlock:Release() - release an instance of a broker block view, recycling it for future use
-- View_BrokerBlock:Configure() - configure the broker block based on current defaults
-- View_BrokerBlock:Update() - update the broker block

-- View attributes:
-- defaults - table with default settings, including those for each class of widget [c]
-- button - widget template for button [c]
-- text - widget template for text field [c]
-- offsetX - horizontal offset from UIParent [c]
-- offsetY - vertical offset from UIParent [c]
-- anchorPoint - if anchor defined then point on UIParent to position relative to (default: BOTTOMLEFT) [c]
-- relativePoint - point on this widget to position relative to (default: BOTTOMLEFT) [c]
-- depth - display depth for this view [c]
-- scale - factor to scale view's display size [c]
-- alpha - alpha for the broker block [c]
-- hasIcon - true if the broker should show the icon
-- hasText - true if the broker block should show the text field
-- hidden - hide the view [d]

local View_BrokerBlock = {}
local EF = Puffin.ErrorFormat
local DB, VC
local classDepth = 0 -- mininum depth for view's of this class

-- Return an initialized instance of the view
function View_BrokerBlock.Create() return VC.CreateView() end

-- Release a view when no longer needed, recycle after clearing attributes and hiding widgets associated with it
function View_BrokerBlock:Release() VC.ReleaseView(self) end

-- Handle mouse interactions with the broker block view
local function BrokerBlock_OnEnter(frame, eventHandler, ...)
	if not eventHandler then return end
	local source, broker = eventHandler.source, eventHandler.broker
	if source == "broker" and broker then
		local _, x, y = frame:GetHitRectInsets()
		-- print("onEnter", broker.label or broker.text or "unknown", x, y, broker.tooltip, broker.OnEnter, broker.OnTooltipShow)
		if broker.tooltip then
			GameTooltip:SetOwner(frame, "ANCHOR_TOPRIGHT", -x, 0)
			if broker.tooltiptext then broker.tooltip:SetText(broker.tooltiptext) end
			broker.tooltip:Show()
		elseif broker.OnEnter then
			GameTooltip:SetOwner(frame, "ANCHOR_TOPRIGHT", -x, 0)
			broker.OnEnter(frame, ...)
			if frame.tooltip then
				frame.tooltip:ClearAllPoints()
				frame.tooltip:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", -x, 0)
			else
				GameTooltip:ClearAllPoints()
				GameTooltip:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", -x, 0)
				GameTooltip:Show()
			end
		elseif broker.OnTooltipShow then
			GameTooltip:SetOwner(frame, "ANCHOR_TOPRIGHT", -x, 0)
			broker.OnTooltipShow(GameTooltip)
			GameTooltip:Show()
		else
			local text = broker.text or broker.label
			if broker.tooltiptext then text = broker.tooltiptext end
			GameTooltip:SetOwner(frame, "ANCHOR_TOPRIGHT", -x, 0)
			GameTooltip:SetText(text or "Unknown")
			GameTooltip:Show()
		end
	end
end

local function BrokerBlock_OnLeave(frame, eventHandler, ...)
	if not eventHandler then return end
	local source, broker = eventHandler.source, eventHandler.broker
	if source == "broker" and broker then
		-- print("onLeave", broker.label or broker.text or "unknown")
		if broker.tooltip then
			GameTooltip:Hide()
		elseif broker.OnLeave then
			broker.OnLeave(frame, ...)
			GameTooltip:Hide()
		else
			GameTooltip:Hide()
		end
	end
end

local function BrokerBlock_OnClick(frame, eventHandler, ...)
	if not eventHandler then return end
	local source, broker = eventHandler.source, eventHandler.broker
	if source == "broker" and broker and broker.OnClick then broker.OnClick(frame, ...) end
end

-- Configure all widgets associated with this view
function View_BrokerBlock:Configure()
	assert(self.defaults, EF("View_BrokerBlock", "undefined defaults"))
	local offsetDepth = ((self.depth or 0) + classDepth) * 10

	if not self._widgets.button and self.button then
		self.button.offsetX = self.offsetX; self.button.offsetY = self.offsetY
		local t = VC.CreateWidget(self.button)
		if t then
			self._widgets.button  = t
			t.onEnter = BrokerBlock_OnEnter; t.onLeave = BrokerBlock_OnLeave; t.onClick = BrokerBlock_OnClick; t.masque = "Broker"
			VC.ConfigureWidget(self, t, self.button, offsetDepth)
		end
	end

	if not self._widgets.text and self.text and self.hasText then
		self.text.width = self.textWidth; self.text.offsetX = self.hasIcon and (self.iconSize + 3) or 0; self.text.offsetY = 0
		self.text.anchor = self._widgets.button ; self.text.anchorPoint = "BOTTOMLEFT"; self.text.relativePoint = "BOTTOMLEFT"
		local t = VC.CreateWidget(self.text)
		if t then
			self._widgets.text = t
			VC.ConfigureWidget(self, t, self.text, offsetDepth)
		end
	end	
end

-- Update all widgets associated with this view, using data from the broker
function View_BrokerBlock:Update(broker)
	local hide = self.hidden
	local t = self._widgets.text
	if t then
		t.hidden = hide
		if self.text and self.hasText then t.text = broker.text else t.text = nil end
		t._class.Update(t)
	end
	t = self._widgets.button 
	if t then
		t.hidden = hide
		if self.button and self.hasIcon then t.texture = broker.icon else t.texture = nil end
		if t.eventHandler then t.eventHandler.source = "broker"; t.eventHandler.broker = broker end
		t._class.Update(t)
	end
end

function View_BrokerBlock.OnEnable()
	DB = Puffin.Component("Data") -- link to data management functions
	assert(DB, EF("View_BrokerBlock", "Data component not found"))
	VC = Puffin.Component("View_Class") -- link to view class functions
	assert(VC, EF("View_BrokerBlock", "View class component not found"))
end

Puffin.RegisterComponent("View_BrokerBlock", View_BrokerBlock, "VIEW", 1)
