-- Puffin: A Tojaso FowlPlay Project
-- View component for cast bars consisting of icon, timer bar, name, timer text.

-- View API:
-- View_CastBar.Create() - a new instance of a broker block view
-- View_CastBar:Release() - release an instance of a broker block view, recycling it for future use
-- View_CastBar:Configure() - configure the broker block based on current defaults
-- View_CastBar:Update() - update the broker block

-- View attributes:
-- defaults - table with default settings [c]
-- layout - table with templates for each class of widget [c]
-- anchorFrame - frame to anchor the icon group to (default: UIParent) [c]
-- anchorPoint - if anchor defined then point on UIParent to position relative to (default: BOTTOMLEFT) [c]
-- relativePoint - point on this widget to position relative to (default: BOTTOMLEFT) [c]
-- offsetX - horizontal offset from anchor [c]
-- offsetY - vertical offset from anchor [c]
-- depth - display depth for this view [c]
-- scale - factor to scale view's display size [c]
-- alpha - alpha for the broker block [c]
-- hidden - hide the view [d]

-- Layout attributes:
-- button - widget template for button [c]
-- bar - widget template for the timer bar [c]
-- label - widget template for label text field [c]
-- timer - widget template for timer text field [c]

local View_CastBar = {}
local EF = Puffin.ErrorFormat
local DB, VC
local classDepth = 0 -- mininum depth for view's of this class

-- Return an initialized instance of the view
function View_CastBar.Create() return VC.CreateView() end

-- Release a view when no longer needed, recycle after clearing attributes and hiding widgets associated with it
function View_CastBar:Release() VC.ReleaseView(self) end

-- Configure all widgets associated with this view
-- Button is primary widget, bar anchors to button, label and timer texts anchor to bar
function View_CastBar:Configure()
	assert(self.defaults, EF("View_CastBar", "undefined defaults"))
	local offsetDepth = ((self.depth or 0) + classDepth) * 10
	local w = self._widgets
	local layout = self.layout
	local dir = (self.iconSide == "left") and 1 or -1
	local bw = self.width - self.iconSize - self.iconOffset -- actual bar width
	if not w.button and layout.button then
		local t = VC.CreateWidget(layout.button)
		if t then
			w.button = t; t.masque = "Castbar"; t.offsetX = self.offsetX; t.offsetY = self.offsetY
			t.iconSize = self.iconSize; t.width = self.height; t.height = self.height
			VC.ConfigureWidget(self, t, layout.button, offsetDepth)
		end
	end
	if not w.bar and layout.bar then
		local t = VC.CreateWidget(layout.bar)
		if t then
			w.bar = t; t.anchor = w.button; t.colors = layout.colors; t.width = bw; t.height = self.height
			if dir > 0 then t.anchorPoint = "LEFT"; t.relativePoint = "RIGHT" else t.anchorPoint = "RIGHT"; t.relativePoint = "LEFT" end
			t.offsetX = self.iconOffset * dir; t.offsetY = 0
			VC.ConfigureWidget(self, t, layout.bar, offsetDepth)
		end
	end	
	if not w.label and layout.label then
		local t = VC.CreateWidget(layout.label)
		if t then
			w.label = t; t.fsize = self.labelSize; t.anchor = w.bar; t.width = bw * 0.75; t.height = self.height
			VC.ConfigureWidget(self, t, layout.label, offsetDepth)
		end
	end	
	if not w.timer and layout.timer then
		local t = VC.CreateWidget(layout.timer)
		if t then
			w.timer = t; t.fsize = self.timerSize; t.anchor = w.bar; t.width = bw * 0.25; t.height = self.height
			VC.ConfigureWidget(self, t, layout.timer, offsetDepth)
		end
	end
	if not w.latency and layout.latency then
		local t = VC.CreateWidget(layout.latency)
		if t then
			w.latency = t; t.anchor = w.bar; t.width = bw - 1; t.height = self.height - 1; t.offsetX = -0.25 * dir; t.offsetY = 0.25
			VC.ConfigureWidget(self, t, layout.latency, offsetDepth)
		end
	end	
end

-- Update all widgets associated with this view, using data in the cast bar v
function View_CastBar:Update(v)
	local hide = self.hidden
	local w = self._widgets
	local t = w.label
	if t then
		t.hidden = hide
		if not hide then t.text = v.label else t.text = nil end
		t._class.Update(t)
	end
	t = w.timer
	if t then
		t.hidden = hide
		if not hide then
			t.timeLeft = v.timeLeft; t.timeMax = v.duration; t.timeExpire = v.expire
		else t.timeMax = nil; t.timeExpire = nil; t.timeLeft = nil end
		t._class.Update(t)
	end
	t = w.button 
	if t then
		t.hidden = hide
		if not hide then t.texture = v.icon else t.texture = nil end
		t._class.Update(t)
	end
	t = w.bar 
	if t then
		t.hidden = hide
		if not hide then
			t.timeLeft = v.timeLeft; t.expire = v.expire; t.duration = v.duration; t.fillBar = not v.channel; t.tickCount = v.ticks
			t.fgColor = v.channel and t.colors.channel or t.colors.cast
			t.bgColor = v.interrupt and t.colors.interrupt or t.colors.background
		else t.expire = nil; t.duration = nil; t.timeLeft = nil end
		t._class.Update(t)
	end
	t = w.latency 
	if t then
		t.hidden = hide
		if not hide and v.latency and v.latency > 0 and v.latency < v.duration then
			t.valueWidth = math.abs(v.latency / v.duration)
		else t.valueWidth = 0 end
		t._class.Update(t)
	end
end

function View_CastBar.OnEnable()
	DB = Puffin.Component("Data") -- link to data management functions
	assert(DB, EF("View_CastBar", "Data component not found"))
	VC = Puffin.Component("View_Class") -- link to view class functions
	assert(VC, EF("View_CastBar", "View class component not found"))
end

Puffin.RegisterComponent("View_CastBar", View_CastBar, "VIEW", 4)
