-- Puffin: A Tojaso FowlPlay Project
-- View component for a display panel that is assembled from an arbitrary collection of widgets, as defined
-- by configuration records containing defaults for each class of widget plus specifics for each widget instance
-- in the view, and updated in response to changes in an assigned data record.

-- Defaults: default attribute values for each class of widget used by the view
-- Widgets: templates that specify exactly which widgets instances go into the view

-- Widget templates include the following attributes:
-- class - widget class name [c]
-- priority - if true then configure this widget before others [c]
-- parent - parent widget for scale and alpha adjustments [c]
-- anchor - widget to anchor to (default: anchor to UIParent plus view's offsetX and offsetY) [c]
-- anchorPoint - if anchor defined then point on anchor to position relative to (default: BOTTOMLEFT) [c]
-- relativePoint - point on this widget to position relative to (default: BOTTOMLEFT) [c]
-- offsetX - horizontal offset from anchor (default: 0) [c]
-- offsetY - vertical offset from anchor (default: 0) [c]
-- width - width of widget [c]
-- height - height of widget [c]
-- scale - factor to scale widget's display size [c]
-- depth - relative to view's depth (range: 0 .. 9) [c]
-- attributes - table of widget-specific attributes applied during configuration [c]
-- providers -- table of data providers used during updates to get attribute values [c]
-- alpha - opacity for the widget [d]
-- hidden - hide the widget [d]

-- View API:
-- View_DisplayPanel.Create() - a new instance of a display panel view
-- View_DisplayPanel:Release() - release an instance of a display panel view, recycling it for future use
-- View_DisplayPanel:Configure() - configure the display panel based on current defaults and template records
-- View_DisplayPanel:Update() - update the display panel

-- View attributes:
-- defaults - table with default settings, including those for each class of widget [c]
-- widgets - table of widget templates actually used to construct this view [c]
-- width - width of the view [c]
-- height - height of the view [c]
-- offsetX - horizontal offset from BOTTOMLEFT of overall screen to BOTTOMLEFT of interactive area (default: 0) [c]
-- offsetY - vertical offset from BOTTOMLEFT of overall screen to BOTTOMLEFT of interactive area (default: 0) [c]
-- depth - display depth for this view [c]
-- scale - factor to scale view's display size [c]
-- hidden - hide the view [d]

local View_DisplayPanel = {}
local EF = Puffin.ErrorFormat
local DB, VC
local classDepth = 0 -- mininum depth for view's of this class

-- Return an initialized instance of the view
function View_DisplayPanel.Create() return VC.CreateView() end

-- Release a view when no longer needed, recycle after clearing attributes and hiding widgets associated with it
function View_DisplayPanel:Release() VC.ReleaseView(self) end

-- Configure all widgets associated with this view
function View_DisplayPanel:Configure()
	assert(self.width and self.width > 0, EF("View_DisplayPanel", "undefined width"))
	assert(self.height and self.height > 0, EF("View_DisplayPanel", "undefined height"))
	assert(self.defaults, EF("View_DisplayPanel", "undefined defaults"))
	assert(self.templates, EF("View_DisplayPanel", "undefined templates"))
	local offsetDepth = ((self.depth or 0) + classDepth) * 10

	local templates = DB.GetRecord("widget_templates", self.templates)
	local widgets = DB.GetRecord("widgets", self.widgets)
	if widgets and templates then
		for i = 1, 10 do -- priority order for configuring (required for parent and position dependencies)
			for _, w in pairs(widgets) do -- each widget has unique name as key into templates table
				local template = templates[w]
				if template and (template.priority == i) then
					local t = self._widgets[w]
					if not t then t = VC.CreateWidget(template); self._widgets[w] = t end
					if template.anchor then t.anchor = self._widgets[template.anchor] else t.offsetX = self.offsetX; t.offsetY = self.offsetY end
					if template.parent then t.parent = self._widgets[template.parent] end
					VC.ConfigureWidget(self, t, template, offsetDepth)
				end
			end
		end
	end
end

-- Update all widgets associated with this view, using widget-specific providers to access the data record
function View_DisplayPanel:Update(record, unit)
	local templates = DB.GetRecord("widget_templates", self.templates)
	local widgets = DB.GetRecord("widgets", self.widgets)
	if widgets and templates then
		for _, w in pairs(widgets) do -- each widget has unique name as key into templates table
			local template = templates[w]
			local t = self._widgets[w]
			assert(t, EF("View_DisplayPanel", "update called for invalid widget"))
			t.hidden = self.hidden
			if not t.hidden and template.providers then -- call providers for widget specific data attributes
				for i, p in pairs(template.providers) do
					if record and type(p) == "table" then
						local v = DB.GetAttribute("providers", p.class, p.name)
						if type(v) == "function" and not p.no_function then v = v(record, unit, p.param) end
						if type(v) == "table" and not p.no_table and v[i] then for j, k in pairs(v) do t[j] = k end else t[i] = v end
					elseif i == "condition" then  -- special case to test for existence of the unit
						t[i] = (p == "exists") and (record ~= nil) or false
					end
				end
			end
			t._class.Update(t)
		end
	end
end

function View_DisplayPanel.OnEnable()
	DB = Puffin.Component("Data") -- link to data management functions
	assert(DB, EF("View_DisplayPanel", "Data component not found"))
	VC = Puffin.Component("View_Class") -- link to view class functions
	assert(VC, EF("View_DisplayPanel", "View class component not found"))
end

Puffin.RegisterComponent("View_DisplayPanel", View_DisplayPanel, "VIEW", 1)
