-- Puffin: A Tojaso FowlPlay Project
-- Invisible widget to handle interactive input as a secure unit frame

-- Custom attributes:
-- unit - unit that the unit frame is assigned [c]

local Widget_Unit = {}
local EF = Puffin.ErrorFormat
local DB
local recycledUnits = {}
local unitCount = 0

-- Manage cache of unit menus that have been "fixed" by removing taint-causing focus items
local menuUnit = nil
local Puffin_UnitFrame_DropDown = nil

-- from a unit, figure out the proper menu and, if appropriate, the corresponding ID
local function LookupMenu(unit)
	if unit == "focus" then return "FOCUS" end
	if UnitIsUnit(unit, "player") then return "SELF" end
	if UnitIsUnit(unit, "vehicle") then return "VEHICLE" end -- must precede pet check since vehicle could also be pet
	if UnitIsUnit(unit, "pet") then return "PET" end	
	if not UnitIsPlayer(unit) then return "TARGET" end
	local id = UnitInRaid(unit); if id then return "RAID_PLAYER", id end
	if UnitInParty(unit) then return "PARTY" end
	return "PLAYER"
end

local function ShowUnitMenu(unit) -- look up menu in cache and, if not there, "fix" it and store in cache
	if menuUnit and UnitPopupMenus then -- set as side effect of the menu event
		local menu, id = LookupMenu(menuUnit)
		if menu and Puffin_UnitFrame_DropDown then UnitPopup_ShowMenu(Puffin_UnitFrame_DropDown, menu, menuUnit, nil, id) end
	end
end

function Unit_Menu(b, unit) menuUnit = unit; ToggleDropDownMenu(1, nil, Puffin_UnitFrame_DropDown, "cursor") end

function Widget_Unit.OnInitialize()
	Puffin_UnitFrame_DropDown = CreateFrame("Frame", "Puffin_UnitFrame_DropDown", UIParent, "UIDropDownMenuTemplate")
	UnitPopupFrames[#UnitPopupFrames+1] = "Puffin_UnitFrame_DropDown"
	UIDropDownMenu_Initialize(Puffin_UnitFrame_DropDown, ShowUnitMenu, "MENU", nil)
end

function Widget_Unit.OnEnable()
	DB = Puffin.Component("Data") -- link to data management functions
	assert(DB, EF("Widget_Unit", "Data component not found"))
end

-- Create a new unit widget object
function Widget_Unit.Create()
	local n, u = next(recycledUnits) -- get any available recycled bar
	if n then
		recycledUnits[n] = nil
	else
		u = {}
		unitCount = unitCount + 1
		u._display = {} -- table of display objects preserved when widget is recycled
		local frameName = "Puffin_Unit" .. tostring(unitCount) -- generate unique name for container frame
		u._display.container = CreateFrame("Button", frameName .. "Container", UIParent, "SecureUnitButtonTemplate")
		u._display.container:SetFrameStrata("MEDIUM")
	end
	u._display.container:Enable()
	return u
end

-- Release when no longer needed, recycle object after clearing attributes and hiding display objects associated with it
function Widget_Unit:Release()
	if InCombatLockdown() then return end -- must not be in combat!
	local b = self._display.container
	if Clique and self._clique then Clique:UnregisterFrame(b) end
	b:EnableMouse(false); b:Disable()
	for k in pairs(self) do if k ~= "_display" then self[k] = nil end end
	table.insert(recycledUnits, self)
end

-- Set display position for the widget relative to another widget or the overall screen
function Widget_Unit:SetPosition(anchor, anchorWidget, relative, offsetX, offsetY, region)
	local t, w = self._display, UIParent
	if anchorWidget then
		if anchorWidget._display then w = region and anchorWidget._display[region] or anchorWidget._display.container else w = anchorWidget end
	end
	t.container:ClearAllPoints(); t.container:SetPoint(anchor, w, relative, offsetX, offsetY)
end

-- Set the parent widget (used for managing display heirarchy)
function Widget_Unit:SetParent(widget)
	local t, w = self._display, widget._display
	if w and w.container then w = w.container else w = UIParent end
	t.container:SetParent(w)
	if not self.depth then t.container:SetFrameLevel(w:GetFrameLevel() + 1) end
end

-- Show tooltip when entering a unit frame
local function Unit_OnEnter(b)
	if GetCVar("UberTooltips") == "1" then
		GameTooltip_SetDefaultAnchor(GameTooltip, b)
	else
		GameTooltip:SetOwner(b, "BOTTOMRIGHT")
	end
	local unit = b.unit
	local t = DB.GetRecord("units", unit)
	if t then if t.inVehicle and t.vehicle then unit = t.vehicle elseif unit == "pet" and t.isPossessed and t.possessor then unit = t.possessor end end
	GameTooltip:SetUnit(unit)
end

-- Hide tooltip when leaving a unit frame
local function Unit_OnLeave() GameTooltip:Hide() end

-- Configure attributes associated with the secure unit button
function Widget_Unit:Configure()
	if InCombatLockdown() then return end -- must not be in combat!
	assert(self.width and self.width > 0, EF("Widget_Unit", "undefined width"))
	assert(self.height and self.height > 0, EF("Widget_Unit", "undefined height"))	
	local b = self._display.container
	local d = self.depth; if d then b:SetFrameLevel(d) end
	b:SetSize(self.width, self.height); b:SetScale(self.scale or 1)
	if self.unit then
		b.unit = self.unit -- used by tooltip code
		b:SetAttribute("unit", self.unit)
		b:SetAttribute("toggleForVehicle", true)
		RegisterUnitWatch(b)
		b:SetScript("OnEnter", Unit_OnEnter)
		b:SetScript("OnLeave", Unit_OnLeave)
		b:EnableMouse(true)
		b:RegisterForClicks("AnyUp")
		b.menu = Unit_Menu
		b:SetAttribute("*type1", "target")
		b:SetAttribute("*type2", "menu")
		b:SetAttribute("*type3", "assist")
	end
	if Clique then -- register this frame with clique if enabled
		if self.clique and not self._clique then
			Clique:RegisterFrame(b); self._clique = true
		elseif not self.clique and self._clique then
			Clique:UnregisterFrame(b); self._clique = nil
		end
	end
end

-- Null update function for unit frame widget
function Widget_Unit:Update() end

-- Register this widget as a component so it can be found by views
Puffin.RegisterComponent("Widget_Unit", Widget_Unit, "WIDGET", 0)
